import datetime

import boto3
import numpy as np
import os
import pandas as pd

from datupapi.configure.config import Config
from datupapi.extract.io import IO
from pandas.tseries.offsets import DateOffset

class Forecasting(Config):

    DOCKER_CONFIG_PATH = os.path.join('/opt/ml/processing/input', 'config.yml')
    io = IO(config_file=DOCKER_CONFIG_PATH, logfile='data_extraction', log_path='output/logs')

    def __init__(self, config_file, logfile, log_path, *args, **kwargs):
        Config.__init__(self, config_file=config_file, logfile=logfile)
        self.log_path = log_path


    def format_forecasts_export(self, df_forecasts):
        """
        Return a dataframe including all forecasts partitions generated by create_forecast_export API's calling

        :param df_forecasts: Dataframe with all contatenated forecasts
        :return df_forecasts: Dataframe concatenating all backtests estimates

        >>> df = format_forecasts_export(df_forecast)
        >>> df =
                        var1    var2    var3
                idx0     1       2       3
        """
        try:
            df_forecasts['date'] = pd.to_datetime(df_forecasts['date'], format='%Y-%m-%d')
            df_forecasts = df_forecasts.rename(columns={'item_id': 'Item',
                                                        'date': 'Date',
                                                        'location': 'Location',
                                                        'p5': 'ForecastLo95',
                                                        'p20': 'ForecastLo80',
                                                        'p40': 'ForecastLo60',
                                                        'p50': 'ForecastPoint',
                                                        'p60': 'ForecastUp60',
                                                        'p80': 'ForecastUp80',
                                                        'p95': 'ForecastUp95'
                                                        }
                                               )
            if self.use_location:
                forecast_cols = [col for col in df_forecasts.columns if all(subcol not in col for subcol in ['Item', 'Date', 'Location'])]
            else:
                forecast_cols = [col for col in df_forecasts.columns if
                                 all(subcol not in col for subcol in ['Item', 'Date'])]
            df_forecasts[forecast_cols] = df_forecasts[forecast_cols].applymap(lambda x: 0 if x < 0 else x)\
                                                                     .applymap(lambda x: round(x))
            df_forecasts['Target'] = 0
            df_forecasts['Week'] = df_forecasts['Date'].dt.isocalendar()['week']
        except KeyError as err:
            self.logger.exception(f'Invalid column name. Please check dataframe metadata: {err}')
            raise
        return df_forecasts


    def compute_errors_from_backtests(self, df_forecast, df_backtest, error_types=['WMAPE', 'MASE']):
        """
        Return a dataframe attaching the specified errors from the averaged out values in backtests

        :param df_forecast: Original dataframe including all items' forecasts
        :param df_backtest: Dataframe including all items' backtests
        :param error_types: List of error to attach to the forecast dataframe. Default WMAPE and MASE
        :return df_forecast: Output forecast dataframe including specified errors

        >>> df = compute_errors_from_backtests(df_forecast, df_backtest, error_types=['WMAPE', 'MASE'])
        >>> df =
                        var1    var2    var3
                idx0     1       2       3
        """
        error_dict = {error_type: 'mean' for error_type in error_types}
        try:
            df_errors = df_backtest.groupby('Item', as_index=True)\
                                   .agg(error_dict)\
                                   .applymap(lambda x: round(x, 2))\
                                   .reset_index(drop=False)
            df_forecast = pd.merge(df_forecast, df_errors, on='Item', how='left')
        except KeyError as err:
            self.logger.exception(f'Invalid column name. Please check dataframe metadata: {err}')
            raise
        return df_forecast


    def compute_interval_per_backtests(self, df_forecast, q_backtest, backtest_names=None, datalake_path_=None):
        """
        Return a dataframe attaching the forecast interval for each backtest dataset

        :param df_forecast: Original dataframe including all items' forecasts
        :param q_backtest: Backtesting dataset name stored in the datalake
        :param backtest_names: List of backtests names to concat, e.g. alpha, bravo, gamma, delta and eps
        :param datalake_path_: Datalake path to download the backtest files from. Do not include bucket name
        :return df_forecast: Output forecast dataframe including specified interval per backtest

        >>> df = compute_interval_per_backtests(df_forecast, backtest_names=['alpha', 'bravo'], datalake_path_='path/tp/backtests/folder')
        >>> df =
                        var1    var2    var3
                idx0     1       2       3
        """
        interval_dict = {}
        try:
            for back_ix, backtest in enumerate(backtest_names):
                df_back = self.io.download_csv(q_name=q_backtest + '-' + backtest, datalake_path=datalake_path_)
                for item in df_back['Item'].unique():
                    mask_back = df_back['Item'] == item
                    mask_fore = df_forecast['Item'] == item
                    interval_dict.update({item: interval for interval in df_back.loc[mask_back, 'SuggestedInterval']})
                    df_forecast.loc[mask_fore, 'SuggIntervalBacktest' + self.backtest_ids[back_ix]] = \
                    df_forecast.loc[mask_fore, 'Item'].map(interval_dict)
            sugg_interval_cols = ['SuggIntervalBacktest' + b for b in self.backtest_ids]
            df_forecast[sugg_interval_cols] = df_forecast[sugg_interval_cols].fillna('Point')
        except KeyError as err:
            self.logger.exception(f'Invalid column name. Please check dataframe metadata: {err}')
            raise
        return df_forecast


    def compute_suggested_interval(self, df_forecast, backtest_names):
        """
        Return a dataframe attaching the suggested interval per date and item record

        :param df_forecast: Original dataframe including all items' forecasts
        :param backtest_names: List of backtests names to concat, e.g. alpha, bravo, etc.
        :param datalake_path_: Datalake path to download the backtest files from. Do not include bucket name
        :return df_forecast: Output forecast dataframe including sugguested interval

        >>> df = compute_suggested_interval(df_forecast, backtest_names=['alpha', 'bravo'])
        >>> df =
                        var1    var2    var3
                idx0     1       2       3
        """
        suggested_interval = []
        num_backtests = len(backtest_names)
        try:
            for ixr, row in df_forecast.iterrows():
                acc_point, acc_lo95, acc_lo80, acc_lo60, acc_up60, acc_up80, acc_up95 = 0, 0, 0, 0, 0, 0, 0
                for back_ix, backtest in enumerate(self.backtest_ids):
                    backtest_factor = num_backtests / (back_ix + 1)
                    if row['SuggIntervalBacktest' + backtest] == 'Point':
                        acc_point += backtest_factor
                    elif row['SuggIntervalBacktest' + backtest] == 'Lo95':
                        acc_lo95 += backtest_factor
                    elif row['SuggIntervalBacktest' + backtest] == 'Lo80':
                        acc_lo80 += backtest_factor
                    elif row['SuggIntervalBacktest' + backtest] == 'Lo60':
                        acc_lo60 += backtest_factor
                    elif row['SuggIntervalBacktest' + backtest] == 'Up60':
                        acc_up60 += backtest_factor
                    elif row['SuggIntervalBacktest' + backtest] == 'Up80':
                        acc_up80 += backtest_factor
                    elif row['SuggIntervalBacktest' + backtest] == 'Up95':
                        acc_up95 += backtest_factor
                suggested_interval_dict = {'Point': acc_point,
                                          'Lo95': acc_lo95,
                                          'Lo80': acc_lo80,
                                          'Lo60': acc_lo60,
                                          'Up60': acc_up60,
                                          'Up80': acc_up80,
                                          'Up95': acc_up95
                                          }
                suggested_interval.append(max(suggested_interval_dict, key=suggested_interval_dict.get))
            df_forecast['SuggestedInterval'] = suggested_interval
        except KeyError as err:
            self.logger.exception(f'Invalid column name. Please check dataframe metadata: {err}')
            raise
        return df_forecast


    def compute_suggested_forecast(self, df_forecast, backtest_names):
        """
        Return a dataframe attaching the suggested interval per date and item record

        :param df_forecast: Original dataframe including all items' forecasts
        :param backtest_names: List of backtests names to concat, e.g. alpha, bravo, etc.
        :param datalake_path_: Datalake path to download the backtest files from. Do not include bucket name
        :return df_forecast: Output forecast dataframe including suggested interval

        >>> df = compute_suggested_forecast(df_forecast, backtest_names=['alpha', 'bravo'])
        >>> df =
                        var1    var2    var3
                idx0     1       2       3
        """
        suggested_interval, suggested_forecast = [], []
        num_backtests = len(backtest_names)
        try:
            for ixr, row in df_forecast.iterrows():
                acc_point, acc_lo95, acc_lo80, acc_lo60, acc_up60, acc_up80, acc_up95 = 0, 0, 0, 0, 0, 0, 0
                for back_ix, backtest in enumerate(self.backtest_ids):
                    backtest_factor = num_backtests / (back_ix + 1)
                    if row['SuggIntervalBacktest' + backtest] == 'Point':
                        acc_point += backtest_factor
                    elif row['SuggIntervalBacktest' + backtest] == 'Lo95':
                        acc_lo95 += backtest_factor
                    elif row['SuggIntervalBacktest' + backtest] == 'Lo80':
                        acc_lo80 += backtest_factor
                    elif row['SuggIntervalBacktest' + backtest] == 'Lo60':
                        acc_lo60 += backtest_factor
                    elif row['SuggIntervalBacktest' + backtest] == 'Up60':
                        acc_up60 += backtest_factor
                    elif row['SuggIntervalBacktest' + backtest] == 'Up80':
                        acc_up80 += backtest_factor
                    elif row['SuggIntervalBacktest' + backtest] == 'Up95':
                        acc_up95 += backtest_factor
                suggested_interval_dict = {'Point': acc_point,
                                          'Lo95': acc_lo95,
                                          'Lo80': acc_lo80,
                                          'Lo60': acc_lo60,
                                          'Up60': acc_up60,
                                          'Up80': acc_up80,
                                          'Up95': acc_up95
                                          }
                suggested_interval.append(max(suggested_interval_dict, key=suggested_interval_dict.get))
                suggested_forecast.append(row['Forecast' + max(suggested_interval_dict, key=suggested_interval_dict.get)])
            df_forecast['SuggestedForecast'] = suggested_forecast
        except KeyError as err:
            self.logger.exception(f'Invalid column name. Please check dataframe metadata: {err}')
            raise
        return df_forecast


    def compute_next_suggested_interval_forecast(self, df_forecast):
        """
        Return a dataframe attaching the next suggested interval and forecast per date and item record

        :param df_forecast: Original dataframe including all items' forecasts
        :return df_forecast: Output forecast dataframe including next suggested interval and forecast

        >>> df = compute_next_suggested_interval_forecast(df_forecast)
        >>> df =
                        var1    var2    var3
                idx0     1       2       3
        """
        intervals_dict = {'0.05': 'Lo95',
                          '0.2': 'Lo80',
                          '0.4': 'Lo60',
                          '0.5': 'Point',
                          '0.6': 'Up60',
                          '0.8': 'Up80',
                          '0.95': 'Up95'
                          }
        next_interval, next_forecast = [], []
        try:
            for ix, row in df_forecast.iterrows():
                if row['SuggestedInterval'] == intervals_dict.get(self.forecast_types[0]):
                    next_interval.append(intervals_dict.get(self.forecast_types[1]))
                    next_forecast.append(row['Forecast' + intervals_dict.get(self.forecast_types[1])])
                elif row['SuggestedInterval'] == intervals_dict.get(self.forecast_types[1]):
                    next_interval.append(intervals_dict.get(self.forecast_types[2]))
                    next_forecast.append(row['Forecast' + intervals_dict.get(self.forecast_types[2])])
                elif row['SuggestedInterval'] == intervals_dict.get(self.forecast_types[2]):
                    next_interval.append(intervals_dict.get(self.forecast_types[3]))
                    next_forecast.append(row['Forecast' + intervals_dict.get(self.forecast_types[3])])
                elif row['SuggestedInterval'] == intervals_dict.get(self.forecast_types[3]):
                    next_interval.append(intervals_dict.get(self.forecast_types[4]))
                    next_forecast.append(row['Forecast' + intervals_dict.get(self.forecast_types[4])])
                elif row['SuggestedInterval'] == intervals_dict.get(self.forecast_types[4]):
                    next_interval.append(intervals_dict.get(self.forecast_types[4]))
                    next_forecast.append(row['Forecast' + intervals_dict.get(self.forecast_types[4])])
                else:
                    next_interval.append(intervals_dict.get(self.forecast_types[2]))
                    next_forecast.append(row['Forecast' + intervals_dict.get(self.forecast_types[2])])
            df_forecast['NextSuggestedInterval'] = next_interval
            df_forecast['NextSuggestedForecast'] = next_forecast
        except KeyError as err:
            self.logger.exception(f'Invalid column name. Please check dataframe metadata: {err}')
            raise
        return df_forecast


    def compute_back_suggested_interval_forecast(self, df_forecast):
        """
        Return a dataframe attaching the back suggested interval and forecast per date and item record

        :param df_forecast: Original dataframe including all items' forecasts
        :return df_forecast: Output forecast dataframe including back suggested interval and forecast

        >>> df = compute_back_suggested_interval_forecast(df_forecast)
        >>> df =
                        var1    var2    var3
                idx0     1       2       3
        """
        intervals_dict = {'0.05': 'Lo95',
                          '0.2': 'Lo80',
                          '0.4': 'Lo60',
                          '0.5': 'Point',
                          '0.6': 'Up60',
                          '0.8': 'Up80',
                          '0.95': 'Up95'
                          }
        back_interval, back_forecast = [], []
        try:
            for ix, row in df_forecast.iterrows():
                if row['SuggestedInterval'] == intervals_dict.get(self.forecast_types[0]):
                    back_interval.append(intervals_dict.get(self.forecast_types[0]))
                    back_forecast.append(row['Forecast' + intervals_dict.get(self.forecast_types[0])])
                elif row['SuggestedInterval'] == intervals_dict.get(self.forecast_types[1]):
                    back_interval.append(intervals_dict.get(self.forecast_types[0]))
                    back_forecast.append(row['Forecast' + intervals_dict.get(self.forecast_types[0])])
                elif row['SuggestedInterval'] == intervals_dict.get(self.forecast_types[2]):
                    back_interval.append(intervals_dict.get(self.forecast_types[1]))
                    back_forecast.append(row['Forecast' + intervals_dict.get(self.forecast_types[1])])
                elif row['SuggestedInterval'] == intervals_dict.get(self.forecast_types[3]):
                    back_interval.append(intervals_dict.get(self.forecast_types[2]))
                    back_forecast.append(row['Forecast' + intervals_dict.get(self.forecast_types[2])])
                elif row['SuggestedInterval'] == intervals_dict.get(self.forecast_types[4]):
                    back_interval.append(intervals_dict.get(self.forecast_types[3]))
                    back_forecast.append(row['Forecast' + intervals_dict.get(self.forecast_types[3])])
                else:
                    back_interval.append(intervals_dict.get(self.forecast_types[2]))
                    back_forecast.append(row['Forecast' + intervals_dict.get(self.forecast_types[2])])
            df_forecast['BackSuggestedInterval'] = back_interval
            df_forecast['BackSuggestedForecast'] = back_forecast
        except KeyError as err:
            self.logger.exception(f'Invalid column name. Please check dataframe metadata: {err}')
            raise
        return df_forecast


    def compute_naive_forecast(self, df_forecast):
        """
        Return a dataframe attaching the back suggested interval and forecast per date and item record

        :param df_forecast: Original dataframe including all items' forecasts
        :return df_forecast: Output forecast dataframe including back suggested interval and forecast

        >>> df = compute_back_suggested_interval_forecast(df_forecast)
        >>> df =
                        var1    var2    var3
                idx0     1       2       3
        """
        try:
            for item in df_forecast['Item'].unique():
                mask = df_forecast['Item'] == item
                target = df_forecast[mask]['Target'].shift(-self.forecast_horizon).fillna(0)
                df_forecast.loc[mask, 'ForecastNaive'] = target
        except KeyError as err:
            self.logger.exception(f'Invalid column name. Please check dataframe metadata: {err}')
            raise
        return df_forecast


    def concat_forecast_backtests(self, df_forecast, df_backtest, debug_cols=False):
        """
        Return a dataframe attaching forecast and backtests records

        :param df_forecast: Original dataframe including all items' forecasts
        :param df_backtest: Dataframe including all items' backtests
        :param debug_cols: True or false to include debug columns from backtest and forecast computation
        :return df_forecast: Output forecast dataframe attaching all backtests

        >>> df = concat_forecast_backtests(df_forecast, df_backtest)
        >>> df =
                        var1    var2    var3
                idx0     1       2       3
        """
        try:
            if debug_cols:
                df_forecast = pd.concat([df_forecast, df_backtest], axis='rows').fillna(0)
            else:
                forecast_cols = [col for col in df_forecast.columns if all(subcol not in col for subcol in ['SuggInterval'])]
                backtest_cols = [col for col in df_backtest.columns if all(subcol not in col for subcol in ['Bias'])]
                df_forecast = pd.concat([df_forecast[forecast_cols], df_backtest[backtest_cols]], axis='rows')
            df_forecast['Date'] = pd.to_datetime(df_forecast['Date'], format='%Y-%m-%d', utc=True)
            df_forecast = df_forecast.sort_values('Date', ascending=False)
        except KeyError as err:
            self.logger.exception(f'Invalid column name. Please check dataframe metadata: {err}')
            raise
        return df_forecast


    def concat_forecast_ranking(self, df_forecast, df_ranking):
        """
        Return a dataframe attaching forecast and item ranking

        :param df_forecast: Original dataframe including all items' forecasts
        :param df_ranking: Dataframe including items ranking
        :return df_forecast: Output forecast dataframe attaching item ranking

        >>> df = concat_forecast_ranking(df_forecast, df_ranking)
        >>> df =
                        var1    var2    var3
                idx0     1       2       3
        """
        try:
            rank_dict = {item: df_ranking.loc[df_ranking['Item']==item, 'Ranking'].unique()[0] for item in df_ranking['Item']}
            df_forecast['Ranking'] = df_forecast['Item'].map(rank_dict)
        except KeyError as err:
            self.logger.exception(f'Invalid column name. Please check dataframe metadata: {err}')
            raise
        return df_forecast


    def make_forecast_strategy(self, df_forecast, item_col, use_location=False):
        """
        Return a dataframe with the strategy forecast analysis per item, including ranking, WMAPE, MASE and interval

        :param df_forecast: Dataframe with forecasts per dates and items
        :param item_col: Item column that identify each reference
        :param use_location: True or False whether the location dimension is included
        :return df_strategy: Dataframe with strategy forecast analysis

        >>> df = make_forecast_strategy(df_forecast, item_col='Item')
        >>> df =
                        var1    var2    var3
                idx0     1       2       3
        """
        if self.dataset_frequency == 'D':
            df_forecast = df_forecast.loc[df_forecast['Date'] >= df_forecast['Date'].max() - DateOffset(days=self.forecast_horizon)]
        elif self.dataset_frequency == 'W':
            df_forecast = df_forecast.loc[df_forecast['Date'] >= df_forecast['Date'].max() - DateOffset(weeks=self.forecast_horizon)]
        elif self.dataset_frequency == 'M':
            df_forecast = df_forecast.loc[df_forecast['Date'] >= df_forecast['Date'].max() - DateOffset(months=self.forecast_horizon)]
        elif self.dataset_frequency == 'Y':
            df_forecast = df_forecast.loc[df_forecast['Date'] >= df_forecast['Date'].max() - DateOffset(years=self.forecast_horizon)]
        else:
            self.logger.debug(f'No valid dataset frequency. Please check config parameters in config.yml')
        try:
            if use_location:
                df_strategy = df_forecast.groupby([item_col], as_index=False).agg({'Location': 'first',
                                                                                   'Ranking': 'first',
                                                                                   'WMAPE': 'mean',
                                                                                   'MASE': 'mean',
                                                                                   'SuggestedInterval': 'first'
                                                                                   })
            else:
                df_strategy = df_forecast.groupby([item_col], as_index=False).agg({'Ranking': 'first',
                                                                                   'WMAPE': 'mean',
                                                                                   'MASE': 'mean',
                                                                                   'SuggestedInterval': 'first'
                                                                                  })
            df_strategy['WMAPE'] = df_strategy['WMAPE'].apply(lambda x: round(x, 2))
            df_strategy['MASE'] = df_strategy['MASE'].apply(lambda x: round(x, 2))
        except KeyError as err:
            self.logger.exception(f'Invalid column name. Please check dataframe metadata: {err}')
            raise
        return df_strategy

